/* sproxy.c: Spy proxy server */
/* This software is released under the GPL, see GPL.txt for more info */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>
#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <curl/curl.h>
#include <curl/easy.h>

#ifndef DEFAULT_PORT
#define DEFAULT_PORT 8080
#endif

#define VERSION "sproxy 1.0"

static FILE *logfp;
#define LOGLEVEL_REQUEST 0x01
#define LOGLEVEL_HEADERS 0x02
#define LOGLEVEL_CONTENT 0x04
static int loglevel = LOGLEVEL_REQUEST | LOGLEVEL_HEADERS;


static void logmsg(const char *msg)
{
    time_t t = time(NULL);
    struct tm *curr_time = localtime(&t);
    char loctime[80];

    strcpy(loctime, asctime(curr_time));
    loctime[strlen(loctime) - 1] = '\0';
    fprintf(logfp, "--- %s: pid %d: ", loctime, getpid());
    fputs(msg, logfp);
    fflush(logfp);
}


static volatile int sigpipe, sigterm;

static void sigpipe_handler(int sig)
{
    sigpipe = 1;
}


static void sigterm_handler(int sig)
{
    char logbuf[64];
    snprintf(logbuf, 64, "Got signal %d, terminating\n", sig);
    logmsg(logbuf);
    sigterm = 1;
}


char *get_request(int sock)
{
    int reqbufsize = 4096;
    char *reqbuf = malloc(reqbufsize + 1);
    int reqbufptr = 0;

    reqbuf[0] = '\0';
    while (1) {
	/* Read data */
	int read = recv(sock, reqbuf + reqbufptr, reqbufsize - reqbufptr, 0);
	if (read <= 0) {
	    if (read < 0) {
		perror("recv");
		free(reqbuf);
		return NULL;
	    }
	    logmsg("Connection closed by client\n");
	    break;
	}
	/* Advance pointer and terminate buffer with \0 */
	reqbufptr += read;
	reqbuf[reqbufptr] = '\0';
	if (reqbufptr > 3) {
	    /* Check if request is complete */
	    if ((strncasecmp(reqbuf, "GET", 3) == 0) &&
		    ((reqbuf[reqbufptr - 2] == '\n' && reqbuf[reqbufptr - 1] == '\n') ||
		    ((reqbuf[reqbufptr - 4] == '\r' && reqbuf[reqbufptr - 3] == '\n' &&
		     reqbuf[reqbufptr - 2] == '\r' && reqbuf[reqbufptr - 1] == '\n')))) {
		/* GET-Request and buffer ends with two newlines: ready */
		break;
	    }
	    if (strncasecmp(reqbuf, "POST", 4) == 0) {
		/* POST-Request: search for Content-Length Header: */
		char *p = strstr(reqbuf, "Content-Length:");
		if (!p) {
		    p = strstr(reqbuf, "content-length:");
		    if (!p) {
			p = strstr(reqbuf, "CONTENT-LENGTH:");
		    }
		}
		if (p) {
		    /* Found Content-Length Header, look for two newlines
		       separating headers and data */
		    int l = atoi(p + 15);
		    int o = 2;
		    char *q = strstr(p, "\n\n");
		    if (!q) {
			o = 4;
			q = strstr(p, "\r\n\r\n");
		    }
		    if (q && strlen(q + o) >= l) {
			/* Data seems to bee complete: ready */
			break;
		    }
		}
	    }
	}
	if (reqbufsize - reqbufptr < 1024) {
	    /* Only space for less than 1024 chars available: enlarge buffer */
	    reqbufsize *= 2;
	    reqbuf = realloc(reqbuf, reqbufsize + 1);
	}
    }
    if (loglevel & LOGLEVEL_REQUEST) {
	fputs("\n", logfp);
	logmsg("Received request from client:\n");
	fputs(reqbuf, logfp);
    }
    return reqbuf;
}


static int send_status = 0;  /* 0: Start, 1: Headers, 2: Body */

static int send_all(char *data, size_t size, size_t num, int *sock)
{
    int count = size * num;
    size = count;

    /* This is an ugly one: Since curl decodes chunked encoded contents,
       we omit the "Transfer-Encoding: chunked" header and send the
       content in one stream. Otherwise we had to encode the content
       again... */
    if (!strncasecmp(data, "Transfer-Encoding: chunked", 26)) {
	return count;
    }
    if (send_status == 0) {
	if ((loglevel & LOGLEVEL_HEADERS) ||
		(loglevel & LOGLEVEL_CONTENT)) {
	    logmsg("Received answer from server:\n");
	}
	send_status = 1;
    }
    if (((send_status == 1) && (loglevel & LOGLEVEL_HEADERS)) ||
	    ((send_status == 2) && (loglevel & LOGLEVEL_CONTENT))) {
	write(fileno(logfp), data, size);
    }
    if (*data == '\r' && *(data + 1) == '\n' && size == 2 && send_status == 1) {
	/* empty line: end of headers */
	send_status = 2;
    }

    do {
	int written = send(*sock, data, size, 0);
	if (written < 0) {
	    if (sigpipe) {
		sigpipe = 0;
		signal(SIGPIPE, sigpipe_handler);
		logmsg("Connection closed by client while sending document\n");
	    } else {
		perror("send");
	    }
	    return -1;
	}
	data += written;
	size -= written;
    } while (size);

    return count;
}


void send_error(int sock, char *msg)
{
    char *head =
    "HTTP/1.1 500 Sproxy Error\r\n"
    "Server: " VERSION "\r\n"
    "Connection: close\r\n"
    "Content-Type: text/html\r\n"
    "\r\n"
    "<HTML>\n<HEAD><TITLE>500 Sproxy Error</TITLE></HEAD>\n"
    "<BODY>\n<H1>Sproxy Error</H1>\n";
   
    char *foot =
    "\n<P><HR><ADDRESS>" VERSION "</ADDRESS>\n"
    "</BODY>\n</HTML>\n";

    send_all(head, strlen(head), 1, &sock);
    send_all(msg, strlen(msg), 1, &sock);
    send_all(foot, strlen(foot), 1, &sock);

    logmsg(msg);
}


static int do_request(int sock, char *req, int ssl)
{
    CURL *curl;
    char errbuf[CURL_ERROR_SIZE];
    struct curl_slist *headers = NULL;
    char *url = NULL;

    /* Init curl session and set some params */
    curl = curl_easy_init();
    curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1);
    // curl_easy_setopt(curl, CURLOPT_MUTE, 1);
    // curl_easy_setopt(curl, CURLOPT_VERBOSE, 1);
    curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, errbuf);
    curl_easy_setopt(curl, CURLOPT_HEADER, 1);
    /* Tell writefunction send_all() to start at the beginning... */
    send_status = 0;
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, send_all);
    /* socket to pass to send_all() */
    curl_easy_setopt(curl, CURLOPT_WRITEDATA, &sock);

    /* Separate Request into lines, and pass the URL and headers to curl */
    if ((strncasecmp(req, "GET ", 4) == 0) ||
	    (strncasecmp(req, "POST ", 5) == 0)) {
	/* First, find the url */
	int post = strncasecmp(req, "POST ", 5) == 0;
	char *q = req + (post ? 5 : 4);
	char *p = strstr(q, " HTTP/");
	if (p) {
	    int err = 0;
	    *p = '\0';
	    if (ssl && strncasecmp(q, "https", 5)) {
		/* replace "http" with "https" in url. */
		url = malloc(strlen(q) + 2);
		strcpy(url, "https");
		strcat(url, q + 4);
	    } else {
		url = q;
	    }
	    curl_easy_setopt(curl, CURLOPT_URL, url);
	    p++;
	    q = strstr(p, "\r\n");
	    if (!q) {
		send_error(sock, "Malformed request, CRLF not found");
		err = 1;
	    }
	    /* And then the headers */
	    while (!err) {
		q += 2;
		if (*q == '\r' && *(q + 1) == '\n') {
		    /* Second CRLF found: end of headers */
		    q += 2;
		    break;
		}
		if (*q) {
		    /* q now points to the first char of a headerline */
		    p = strstr(q, "\r\n");
		    if (p) {
			*p = '\0';
			headers = curl_slist_append(headers, q);
			q = p;
		    } else {
			send_error(sock, "Malformed request, CRLF not found");
			err = 1;
			break;
		    }
		} else {
		    send_error(sock, "Malformed request, no header found after single CRLF");
		    err = 1;
		    break;
		}
	    }
	    if (!err) {
		/* We have parsed the headers. q now points to the data
		   to post, if any */
		curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);
		if (post) {
		    curl_easy_setopt(curl, CURLOPT_POSTFIELDS, q);
		}
		err = curl_easy_perform(curl);
		if (err) {
		    send_error(sock, errbuf);
		}
	    }
	} else {
	    send_error(sock, "Malformed request, protocol not found");
	}
    } else {
	send_error(sock, "Can only handle GET and POST requests");
    }

    /* Cleanup before leaving... */
    curl_easy_cleanup(curl);
    curl_slist_free_all(headers);
    if (url) free(url);
    return 0;
}


static void usage(const char *me)
{
    fprintf(stderr, "Usage: %s [ OPTIONS ]\n", me);
    fprintf(stderr, "-p NUM --port=NUM\n\taccept requests on port NUM (default %d)\n", DEFAULT_PORT);
    fprintf(stderr, "-l FILE --logfile=FILE\n\tlog requests to file FILE (default stdout)\n");
    fprintf(stderr, "-L LOGLEVEL --loglevel=LOGLEVEL\n\tset loglevel (default 3):\n"
	"\t0 = no logging, 1 = request, 2 = headers, 3 = request and headers\n"
	"\tone of the above + 4 = log also the content\n");
    fprintf(stderr, "-f NUM --fork=NUM\n\tfork NUM server processes (default 0)\n");
    fprintf(stderr, "-s --ssl\n\tuse ssl (https) to connect to the server\n"
	"\teven if client requests only http\n");
    fprintf(stderr, "-h --help\n\tthis screen\n");
    exit(1);
}


int main(int argc, char *argv[])
{
    struct sockaddr_in me;
    int sock, msgsock, flag;
    unsigned short port = DEFAULT_PORT;
    char *logfile = NULL;
    char logbuf[128];
    int fork_n = 0, n_chld = 0;
    int ssl = 0;
    pid_t *children = NULL;
    int c, longind;
    struct option longopts[] = {
	{"port", 1, NULL, 'p'},
	{"logfile", 1, NULL, 'l'},
	{"loglevel", 1, NULL, 'L'},
	{"fork", 1, NULL, 'f'},
	{"ssl", 0, NULL, 's'},
	{"help", 0, NULL, 'h'},
	{NULL, 0, NULL, 0}
    };

    while ((c = getopt_long(argc, argv, "p:l:L:f:sh", longopts, &longind)) != EOF) {
	switch (c) {
	case 'p':
	    port = atoi(optarg);
	    break;
	case 'l':
	    logfile = optarg;
	    break;
	case 'L':
	    loglevel = atoi(optarg);
	    break;
	case 'f':
	    fork_n = atoi(optarg);
	    break;
	case 's':
	    ssl = 1;
	    break;
	case 'h':
	case '?':
	    usage(argv[0]);
	    return 0;
	}
    }

    if (fork_n) {
	children = (pid_t *) calloc(fork_n, sizeof(pid_t));
	if (!children) {
	    fprintf(stderr, "%s: memory exhausted\n", argv[0]);
	    exit(1);
	}
    }

    if (logfile) {
	logfp = fopen(logfile, "a");
	if (!logfp) {
	    perror(logfile);
	    exit(1);
	}
    } else {
	logfp = stdout;
    }

    signal(SIGPIPE, sigpipe_handler);
    signal(SIGINT, sigterm_handler);
    signal(SIGTERM, sigterm_handler);

    siginterrupt(SIGPIPE, 1);
    siginterrupt(SIGINT, 1);
    siginterrupt(SIGTERM, 1);

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	perror("opening stream socket");
	fclose(logfp);
	exit(1);
    }

    flag = 1;
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (const void *) &flag,
	    sizeof(int)) < 0) {
	perror("setsockopt(SO_REUSEADDR)");
    }

    me.sin_family = AF_INET;
    me.sin_addr.s_addr = INADDR_ANY;
    me.sin_port = htons(port);
    if (bind(sock, (struct sockaddr *) &me, sizeof me) < 0) {
	perror("binding stream socket");
	fclose(logfp);
	exit(1);
    }

    /* start accepting connections */
    listen(sock, 5);

    snprintf(logbuf, 128, "%s listening on port %u\n", VERSION, port);
    logmsg(logbuf);

    if (fork_n) {
	for (n_chld = 0; n_chld < fork_n; n_chld++) {
	    children[n_chld] = fork();
	    if (children[n_chld] < 0) {
		perror("fork");
		break;
	    } else if (children[n_chld] == 0) {
		free(children);
		children = NULL;
		logmsg("Server process starting\n");
		break;
	    }
	}
    }

    if (children) {
	int i;

	sprintf(logbuf, "Forked %d server processes\n", n_chld);
	logmsg(logbuf);

	while (!sigterm) {
	    sleep(60);
	}

	for (i = 0; i < n_chld; i++) {
	    if (waitpid(children[i], NULL, 0) < 0) {
		perror("waitpid");
		sprintf(logbuf, "waitpid() failed for child process %d\n", children[i]);
		logmsg(logbuf);
	    }
	}
	logmsg("All child processes terminated, exiting\n");
    } else {
	while (!sigterm) {
	    char *req;

	    msgsock = accept(sock, NULL, NULL);

	    if (msgsock == -1) {
		if (sigterm) {
		    break;
		}
		/* perror("accept"); */
		continue;
	    }

	    /* This will malloc() a buffer and store the request into it */
	    req = get_request(msgsock);
	    if (req) {
		do_request(msgsock, req, ssl);
		/* Free the memory allocated by get_request() above */
		free(req);
	    }

	    close(msgsock);
	}
    }

    close(sock);
    if (logfile) fclose(logfp);

    return 0;
}

